import 'package:matrix/matrix.dart';
import 'package:moor/moor.dart';

class Server {
  String serverUrl = 'https://matrix.org';
  String? cName = 'client';
  final String middleUrl = '_matrix/client/r0';
  Client client = Client('client');
  dynamic call;
  String clientSyncState = '';
  bool isAuthenticated = false;
  LoginResponse? loginResponse;
  late PublicRoom broadcastAdHocMessagesRoom;

  set broadcastAdhocMessageRoom(PublicRoom room) {
    broadcastAdHocMessagesRoom = room;
  }

  String get userId => client.userID;

  Server() {
    client.checkHomeserver(serverUrl);
  }

  Server.init(String url, String? name)
      : serverUrl = url,
        cName = name {
    client = Client(cName);
    client.checkHomeserver(serverUrl);
  }

  @override
  String toString() {
    return '$serverUrl-$cName';
  }

  void getLoginInfo() {
    return;
  }

  // it logs in automatically..
  Future<LoginResponse> registration(username, password) async {
    return client.register(username: username, password: password);
  }

  Future<LoginResponse> login(username, password) async {
    return client
        .login(
      // user: username,
      password: password,
      identifier: AuthenticationIdentifier.subFromJson(
          {'type': 'm.id.user', 'user': username}),
    )
        .then((response) async {
      await setBroadcastingAdHocMessagesRoom();
      isAuthenticated = true;
      return response;
    });
  }

  // server is forced to maintain only 1 public room
  Future<void> setBroadcastingAdHocMessagesRoom() async {
    PublicRoomsResponse room = await client.getPublicRooms(limit: 1);
    broadcastAdHocMessagesRoom = room.chunk[0];
    Room? roomInfo = client.getRoomById(room.chunk[0].roomId);
    if (roomInfo == null || roomInfo.membership == Membership.invite) {
      await joinRoom(room.chunk[0].roomId);
    }
  }

  Future<Visibility> getVisibilityOfRoom(String roomId) {
    return client.getRoomVisibilityOnDirectory(roomId);
  }

  Future<List<List<MatrixEvent>>> getJoinedRooms() async {
    return client.getJoinedRooms().then((rooms) {
      rooms.retainWhere((e) => e != broadcastAdHocMessagesRoom.roomId);
      return Future.wait(rooms.map((room) async {
        List<MatrixEvent> member =
            await client.getMembersByRoom(room, notMembership: Membership.ban);
        return member;
      }));
    });
  }

  BasicEvent? fetchAccountData(String eventType) {
    return client.accountData[eventType];
  }

  String getAccessToken() {
    return client.accessToken;
  }

  /// creates a room or return existing room
  Future<String> createRoom(
      {String? roomAliasName,
      String? name,
      String? topic,
      List<String>? invite}) async {
    print('creating a new room');
    final userId = invite?[0];
    String roomId = client.getDirectChatFromUserId(userId);
    if (roomId != null) {
      return Future.value(roomId);
    }
    return client
        .createRoom(
      visibility: Visibility.private,
      preset: CreateRoomPreset.private_chat,
      isDirect: true,
      roomAliasName: roomAliasName ?? '',
      name: name ?? '',
      topic: topic ?? '',
      invite: invite ?? [],
    )
        .then((value) {
      client.setAccountData(client.userID, 'm.direct', {
        userId.toString(): [value]
      });
      return value;
    });
  }

  List<Room> getInvites() {
    return client.rooms
        .where((room) => room.membership == Membership.invite)
        .toList();
  }

  Future<void> inviteToRoom(String roomId, String userId) async {
    return client.inviteToRoom(roomId, userId);
  }

  Future<String> joinRoom(String roomId) async =>
      client.joinRoomById(roomId).then((roomId) async {
        List<MatrixEvent> members =
            await client.getMembersByRoom(roomId, membership: Membership.join);
        await Future.wait(members
            .map((member) => client.setAccountData(client.userID, 'm.direct', {
                  member.senderId.toString(): [roomId]
                }))
            .toList());
        return roomId;
      });

  Future<PresenceContent> userPresence(String userId) async {
    return client.getPresence(userId);
  }

  Future<String> sendMessage(String userId, String message,
      {String? roomId}) async {
    roomId ??= await createRoom(invite: [
      userId,
    ]);
    var room = client.getRoomById(roomId);
    try {
      final presence = await userPresence(userId);
      if (presence.presence == PresenceType.offline) {
        await sendAdHocBroadcastMessage(message);
      }
      return room.sendTextEvent(message);
      // return client.sendMessage(roomId, 'm.room.message', txnId, content);
    } on Exception {
      await sendAdHocBroadcastMessage(message);
      return room.sendTextEvent(message);
      // return client.sendMessage(roomId, 'm.room.message', txnId, content);
    }
  }

  Future<String> sendAdHocBroadcastMessage(String message) async {
    // String txnId = client.generateUniqueTransactionId();
    // final content = {
    //   'msgtype': 'm.text',
    //   'body': message,
    // };
    var room = client.getRoomById(broadcastAdHocMessagesRoom.roomId);
    var msg = await room.sendTextEvent(message);
    print(
        'event id: $msg, to public room id: ${broadcastAdHocMessagesRoom.roomId}');
    return msg;
    // return client.sendMessage(
    //     broadcastAdHocMessagesRoom.roomId, 'm.room.message', txnId, content);
  }

  Future<void> logout() {
    return client.logout();
  }

  PresenceType myPresence() {
    return client.syncPresence;
  }

  void onEvent(void Function(EventUpdate) callback) {
    client.onEvent.stream.listen(callback);
  }

  void on(String event, void Function(dynamic) callback) {
    switch (event) {
      case 'sync':
        client.onSync.stream.listen(callback);
        break;
      case 'presence':
        client.onRoomUpdate.stream.listen(callback);
        break;
      case 'callInvite':
        client.onCallInvite.stream.listen(callback);
        break;
      case 'callAnswer':
        client.onCallAnswer.stream.listen(callback);
        break;
      case 'callCandidates':
        client.onCallCandidates.stream.listen(callback);
        break;
      case 'callHangup':
        client.onCallHangup.stream.listen(callback);
        break;

      default:
        break;
    }
  }
}
